"""
runs a resync function for the product and teask called
"""
from meraki.exceptions import AsyncAPIError
from autosync import lib, model


async def re_sync(sdk: object, org_id: str, net_id: str,
                  task: str, product: str):
    """
         Function reSyces a spefic meraki dashbaord eliment
    Args:
        sdk(object): Meraki dashboard SDK Objeect
        org_id: Meraki Org ID
        net_id(str): Network ID String
        task(str): Task to resync
        product(str): Meraki Product being synced I.E. switch, wirless

    Returns:
       Nothing Updateds the network Object that is passed
    """
    try:
        action = getattr(sdk, product)
        value = getattr(action, f'get{task}')
        setattr(model.meraki_nets[org_id].networks[net_id].dashboard[product],
                task, await value(net_id))
    except AsyncAPIError as apie:
        print(
            f'\t {lib.bc.FAIL} Error Running Setting {task} '
            f'{lib.bc.WARNING}Error Message: {str(apie)}{lib.bc.Default}')
    except Exception as error:
        print(
                f'{lib.bc.WARNING}Error with Module: {str(error)}'
                f'{lib.bc.Default}')



"""
Functions for loading, storing and clearing of Cache Files
This can also be used for backup function in future releases
"""
import pickle
from datetime import datetime
from os import path, remove
from autosync import const, model


def cache_age(item):
	"""
	Determins age of Cache file
	Args:
		item: Data Model item that has the last_sync value

	Returns: time delta from time utc now to the value stored in last+sync

	"""
	return datetime.utcnow() - item.last_sync


# returns the cached object if exists otherwise returns the locally synced
def load_cache(org_id: str, is_golden: bool):
	"""
	Lods Cache into Memory from stored pickle files
	Args:
		is_golden: If Master True Load org from Master Dict
		org_id: Org ID of meraki Org to load

	Returns:

	"""
	if is_golden:
		org = model.golden_nets[org_id]
	else:
		org = model.meraki_nets[org_id]
	cache_file_str = \
		f'{const.appcfg.cache_dir}/{str(org.org_id)}-{str(org.name)}.mnet'
	if path.exists(cache_file_str) and const.appcfg.use_cache:
		with open(cache_file_str, "rb") as cache_file:
			org_cached = pickle.load(cache_file)
		if org_cached.lastsync is None:
			print('Has Cache! But it is stale, re-syncing')
			clear_cache(org_id, is_golden)
			org.cached = False
		elif datetime.utcnow() - datetime.fromisoformat(str(org_cached.lastsync)):
			org.__dict__ = org_cached.__dict__.copy()
		else:
			print('Cache File Error! But it is stale, re-syncing')
			clear_cache(org_id, is_golden)
			org.cached = False
	else:
		print('No Cache Found')
		org.cached = False


# writes it to disk for faster load times
def store_cache(org_id: str, is_golden: bool):
	"""
	Stores Cache to file system as pickle files
	Args:
		org_id: Orginixation ID
		is_golden (object): Golden Network

	Returns:

	"""
	if not const.appcfg.use_cache:
		return
	if is_golden:
		org = model.golden_nets[org_id]
		model.golden_nets[org_id].change_log = []
	else:
		org = model.meraki_nets[org_id]
		model.meraki_nets[org_id].change_log = []
	org.cached = True
	
	clear_cache(org_id, is_golden)
	cache_file = f'{const.appcfg.cache_dir}/{str(org.org_id)}-{str(org.name)}.mnet'
	pickle.dump(org, open(cache_file, "wb"))


# clears cache and kills disk backup
def clear_cache(org_id: str, is_golden: bool):
	"""
	Clears Cache ( Delets Pickle file from File System
	Args:
		org_id: Orginixation ID
		is_golden: Golden Network

	Returns:

	"""
	if is_golden:
		org = model.golden_nets[org_id]
	else:
		org = model.meraki_nets[org_id]
	cache_file_str = \
		f'{const.appcfg.cache_dir}/{str(org.org_id)}-{str(org.name)}.mnet'
	if path.exists(cache_file_str):
		remove(cache_file_str)



"""
Start Meraki API either Async or Sync
"""
import asyncio
import logging
import os
import meraki
import meraki.aio
from autosync import const


class MerakiAsyncApi:
    """
    Meraki Async SDK
    """
    
    def __init__(self):
        self._appcfg = const.appcfg
        self.api = None
    
    def __enter__(self):
        self.api = meraki.aio.AsyncDashboardAPI(
                api_key=self._appcfg.MERAKI_DASHBOARD_API_KEY,
                base_url=self._appcfg.meraki_base_url,
                simulate=self._appcfg.simulate,
                wait_on_rate_limit=self._appcfg.wait_on_rate_limit,
                maximum_concurrent_requests=self._appcfg.maximum_concurrent_requests,
                nginx_429_retry_wait_time=self._appcfg.nginx_429_retry_wait_time,
                maximum_retries=int(self._appcfg.maximum_retries),
                log_file_prefix=os.path.basename(__file__)[:-3],
                log_path=self._appcfg.log_path,
                suppress_logging=self._appcfg.suppress_logging,
                print_console=False)
        
        return self.api
    
    def __exit__(self, type, value, traceback):
        # Exception handling here
        loop = asyncio.get_event_loop()
        loop.create_task(self.api._session.close())
    
    def __del__(self):
        loop = asyncio.get_event_loop()
        loop.create_task(self.api._session.close())
    
    async def closesession(self):
        """
            Cloes session for ASYCN SDK
            Returns:
        """
        await self.api._session.close()


class MerakiApi:
    """
    Meraki Sync API init for SYNC SDK
    """
    
    def __init__(self):
        """
        Init fucntion for Meraki SYNC SDK
        """
   
        self.api = meraki.DashboardAPI(
                api_key=const.appcfg.MERAKI_DASHBOARD_API_KEY,
                base_url=const.appcfg.meraki_base_url,
                simulate=const.appcfg.simulate,
                wait_on_rate_limit=const.appcfg.wait_on_rate_limit,
                log_file_prefix=os.path.basename(__file__)[:-3],
                log_path=const.appcfg.log_path,
                suppress_logging=const.appcfg.suppress_logging)
        logger = logging.getLogger('meraki')
        logger.setLevel(logging.WARNING)
    
    def get_api(self):
        """
        Returms Meraki SDK Sync API Object
        Returns:

        """
        return self.api



import os

import click


@click.group()
def cli():
    """ Tasks to start sync auto Sycn of Meraki Networks """
    pass

@click.command(help='Sets the Meraki API Key as an environment variable')
@click.option('-key','--key', required=True)
def setkey(key):
    os.environ['meraki_dashboard_api_key'] = key

@click.command(help='Display the Meraki API Key as an environment variable ')
def showkey():
    print(os.getenv('meraki_dashboard_api_key','No key set please run autosync setkey --key <API Key> to set your key'))


cli.add_command(setkey)
cli.add_command(showkey)



"""
Meraki Dashboard Switch Settings
"""
import copy
from meraki.exceptions import AsyncAPIError
from autosync import lib, const
from autosync.mnetutils.resync import re_sync


class switch:
    """
        Meraki Dashbaord Data Class for all switch swttings
    """
    def __init__(self):
        self.NetworkSwitchMtu = None
        self.NetworkSwitchSettings = 'Test'
        self.NetworkSwitchDscpToCosMappings = None
        self.NetworkSwitchRoutingMulticast = None
        self.NetworkSwitchAccessControlLists = None
        self.NetworkSwitchStormControl = None
        self.NetworkSwitchQosRules = None
        self.NetworkSwitchQosRulesOrder = None
        self.NetworkSwitchAccessPolicy = None
        self.NetworkName = None

    def settings(self):
        """
        Fucntion to return settings to cycle through when pulling configuration
        from Meraki Dashboard Skipping objects that don't refer to a dashbard
        API function
        Returns: Settings to Cycle Through
        """
        skip = ['NetworkName']
        items = [i for i in self.__dict__.keys() if i not in skip]
        return items

    @staticmethod
    def _apply_rad_keys(_policy):
        """
        Apply Radius Keys to servers for access policy
        Args:
            _policy: Policy DiCT

        Returns:Policy dict with RAD Keys

        """
        for server in _policy['radiusServers']:
            server['secret'] = const.appcfg.rad_keys_all
        for server in _policy['radiusAccountingServers']:
            server['secret'] = const.appcfg.rad_keys_all
        return _policy

    async def Get_NetworkSwitchAccessPolicy(self, sdk, net_id, net_name):
        """
        Gets Switch Access Policy due to Name Difference
        Args:
            sdk: Meraki API Object
            net_id: Network ID
            net_name: Network Name

        Returns:

        """
        if const.appcfg.debug:
            print(f'Network ID: {net_id} - {net_name}')
        self.NetworkSwitchAccessPolicy = await sdk.switch.getNetworkSwitchAccessPolicies(
            net_id)

    @staticmethod
    async def Update_NetworkSwitchQosRulesOrder(sdk: object, org_id: str,
                                                net_id: str, net_name):
        """
        fucntion to update network qos rele order but needs to be written
        Args:
            sdk: Meraki SDK Object
            org_id: Current Org ID
            net_id: Curremt Metwork ID
            net_name: Network Name
        Returns:

        """
        if const.appcfg.debug:
            print(f'ORG ID: {org_id} - Network ID: {net_id} - {net_name}')
        # TODO Need to look at why this does not work and if it is needed moving forward
            print("Not IMplemented")

    async def Update_NetworkSwitchAccessPolicy(self, sdk: object, org_id: str,
                                               net_id: str, net_name):
        """
        Update Switch Access Polcies for 802.1x radius servers
        Args:
            sdk: Meraki SDK Object
            org_id: Current Org ID
            net_id: Curremt Metwork ID
            net_name: Network Name
        Returns:

        """
        if const.appcfg.debug:
            print(f'ORG ID: {org_id} - Network ID: {net_id} - {net_name}')
        golden = lib.get_golden('switch')
        resync = True
        for policy in golden.NetworkSwitchAccessPolicy:
            exitist = False
            update = False
            for net_policy in self.NetworkSwitchAccessPolicy:
                if not await lib.compare(policy, net_policy):
                    exitist = True
                    update = True
                else:
                    update = False
                    exitist = True
            if not exitist:
                resync = True
                policy = self._apply_rad_keys(policy)
                await sdk.switch.createNetworkSwitchAccessPolicy(
                    net_id, **policy)
                lib.print_update(net_name, "Access Policy", "Switch")
            elif update:
                resync = True
                policy = self._apply_rad_keys(policy)
                await sdk.switch.updateNetworkSwitchAccessPolicy(
                    net_id, **policy)
                lib.print_update(net_name, "Access Policy", "Switch")

                lib.print_matched(net_name, "Access Policy", "Switch")
        if resync:
            await self.Get_NetworkSwitchAccessPolicy(sdk, net_id, net_name)
        else:
            lib.print_matched(net_name, "Access Policy", "Switch")

    async def Update_NetworkSwitchAccessControlLists(self, sdk: object,
                                                     org_id: str, net_id: str,
                                                     net_name):
        """
        Update Switch Access Control Lists
        Args:
            sdk: Meraki SDK Object
            org_id: Current Org ID
            net_id: Curremt Metwork ID
            net_name: Network Name
        Returns:

        """
        if const.appcfg.debug:
            print(f'ORG ID: {org_id} - Network ID: {net_id} - {net_name}')
        golden = lib.get_golden('switch')
        if not await lib.compare(golden.NetworkSwitchAccessControlLists,
                                 self.NetworkSwitchAccessControlLists):
            if const.appcfg.write:
                acls = copy.deepcopy(golden.NetworkSwitchAccessControlLists)
                # remove the default rule at the end
                acls['rules'].remove(acls['rules'][len(acls['rules']) - 1])
                lib.print_update(net_name, "Access Control Lists", "Switch")
                await sdk.switch.updateNetworkSwitchAccessControlLists(
                    net_id, **acls)
                await re_sync(sdk, org_id, net_id,
                              'NetworkSwitchAccessControlLists', 'switch')
        else:
            lib.print_matched(net_name, "Access Control Lists", "Switch")

    async def Update_NetworkSwitchQosRules(self, sdk: object, org_id: str,
                                           net_id: str, net_name):
        """
        Updates Network QOS Rules
        Args:
            sdk: Meraki SDK Object
            org_id: Current Org ID
            net_id: Curremt Metwork ID
            net_name: Network Name
        Returns:

        """
        if const.appcfg.debug:
            print(f'ORG ID: {org_id} - Network ID: {net_id} - {net_name}')
        golden = lib.get_golden('switch')
        if not await lib.soft_compare(golden.NetworkSwitchQosRules,
                                      self.NetworkSwitchQosRules):
            # {'ruleIds': ['577586652210270187', '577586652210270188'
            # , '577586652210270189']}
            r_order_src = golden.NetworkSwitchQosRules
            r_order_dst = self.NetworkSwitchQosRules
            qos_runs = 0
            for rid in r_order_src:
                rid_exists = False
                for rid2 in r_order_dst:
                    if rid['vlan'] is None or rid['vlan'] == rid2['vlan']:
                        if rid['protocol'] is None or rid['protocol'] == \
                                rid2['protocol']:
                            if rid['srcPort'] is None or rid['srcPort'] == \
                                    rid2['srcPort']:
                                if rid['dstPort'] is None or rid[
                                        'dstPort'] == rid2['dstPort']:
                                    if rid['dstPort'] is None or rid[
                                            'dscp'] == rid2['dscp']:
                                        rid_exists = True
                                        continue
                if rid_exists:
                    # print(f'Duplicate rule, skipping!')
                    continue
                if qos_runs == 0:
                    qos_runs += 1
                    print(f'\t{lib.bc.OKGREEN}-Cloning Switch QoS Rules...')
                # [{'net_id': '577586652210270187','vlan':
                # None,'protocol': 'ANY','srcPort': None,
                # 'dstPort': None,'dscp': -1}, .. ]
                for qos_rule in golden.NetworkSwitchQosRules:
                    if qos_rule['id'] == rid['id']:
                        rule = copy.deepcopy(qos_rule)
                        # try:
                        # pop the net_id, and srcPort/dstPort if they're empty,
                        # otherwismne it'll throw an error
                        rule.pop('id')
                        if rule['srcPort'] is None:
                            rule.pop('srcPort')
                        if rule['dstPort'] is None:
                            rule.pop('dstPort')
                        try:
                            if const.appcfg.write:
                                await sdk.switch.createNetworkSwitchQosRule(
                                    net_id, **rule)
                                lib.print_update(net_name, "Switch QOS Rules",
                                                 "Switch")
                        except AsyncAPIError as error:
                            print(
                                f'\t {lib.bc.FAIL} Error running api: '
                                f'{lib.bc.WARNING} {str(error)}{lib.bc.Default}'
                            )
                    else:
                        print(
                            f'{lib.bc.FAIL}ERROR FINDING QoS RULE!!!{lib.bc.ENDC}'
                        )
        else:
            lib.print_matched(net_name, "Switch QOS Rules", "Switch")



"""
Proccess functions to loop through networks with in an orginizaion and
send that o
"""
import asyncio

import random
import time
from autosync import mnetutils, model, lib, const
from datetime import datetime


async def validate_network_processor(org_id: str, net_id: str,
                                     sdk: object):
	"""
    ASYNC Function to synclye through configuration dict for network and
    call the clone functions based on netork net_id 
    Args: 
        org_id:Orginization Id 
        net_id: Network Id 
        sdk: Meraki SDK Object

    Returns:

    """
	print(
			f'{lib.bc.OKGREEN}Starting Network configuration validation...{lib.bc.ENDC}'
	)
	for product in model.meraki_nets[org_id].networks[net_id].dashboard:
		await mnetutils.clone(org_id, net_id, sdk, product)


async def proccess_network(org_id: str, net_id: str, sdk: object,
                           tags: list):
	"""
    Function checkts Tag of network for Master tag then loops through
    configuration in network and saves items into the Class objects
	Args:
        org_id:Orginization Id 
        net_id: Network Id 
        sdk: Meraki SDK Objectt

	Returns:

	"""
	t1 = time.perf_counter()
	
	is_golden = await check_is_golden(tags, net_id)
	if is_golden:
		net = model.golden_nets[const.appcfg.tag_golden].networks[
			const.appcfg.tag_golden]
		net_name = const.appcfg.tag_golden
		last_change_ts = await lib.last_change(
			model.golden_nets[const.appcfg.tag_golden].change_log,
			net_id)
		is_changed = await check_last_change(net.change_log_ts,
		                                     last_change_ts)
		if model.golden_nets[
			const.appcfg.tag_golden].cached and is_changed:
			print(
				f'{lib.bc.WARNING} Golden Network Tag: {const.appcfg.tag_golden} Starting Configuration Sync at {t1:0.5f} secound {lib.bc.OKBLUE} - Cached Change TS: '
				f'{model.golden_nets[const.appcfg.tag_golden].networks[const.appcfg.tag_golden].change_log_ts} '
				f'Current TS: {str(last_change_ts)} {lib.bc.ENDC}')
			sync = True
		elif model.golden_nets[
			const.appcfg.tag_golden].cached and not is_changed:
			print(
				f'{lib.bc.OKGREEN} Golden Network Tag: {const.appcfg.tag_golden} is CACHED and has not changed  {lib.bc.ENDC}')
			sync = False
		else:
			print(
					f'{lib.bc.OKGREEN}Started Configuration Sync at {t1:0.5f} secound'
					f' Golden Network Tag: {const.appcfg.tag_golden} is not Cached ')
			sync = True
	else:
		last_change_ts = await lib.last_change(
			model.meraki_nets[org_id].change_log, net_id)
		net = model.meraki_nets[org_id].networks[net_id]
		net_name = model.meraki_nets[org_id].networks[net_id].name
		is_changed = await check_last_change(net.change_log_ts,
		                                     last_change_ts)
		
		if model.meraki_nets[org_id].cached and is_changed:
			print(
					f'{lib.bc.WARNING} Network: {net_name} HAS CHANGED  {lib.bc.OKBLUE} Starting Configuration Sync at {t1:0.5f} secound {lib.bc.OKBLUE} - Cached Change TS: '
					f'{model.meraki_nets[org_id].networks[net_id].change_log_ts} '
					f'Current TS: {str(last_change_ts)} {lib.bc.ENDC}')
			sync = True
		elif model.meraki_nets[org_id].cached and not is_changed:
			print(
					f'{lib.bc.OKGREEN} Network: {net_name} is CACHED and has not changed  {lib.bc.ENDC}')
			sync = False
		else:
			print(
					f'{lib.bc.OKGREEN}Started Configuration Sync at {t1:0.5f} secound'
					f' Network: {net_name} is not Cached ')
			sync = True
	if 'networks' not in net.products:
		net.products.append('networks')
	approvedList = net.supported
	if sync:
		if const.appcfg.tag_target in net.tags or const.appcfg.tag_override:
			for product in net.products:
				if product in approvedList:
					waiting = random.randrange(0, 5)
					await asyncio.sleep(waiting)
					if is_golden:
						await mnetutils.sync(sdk, const.appcfg.tag_golden,
						                     const.appcfg.tag_golden,
						                     product, is_golden)
					else:
						try:
							await mnetutils.sync(sdk, net.org_id,
							                     net.net_id, product,
							                     is_golden)
						except Exception as error:
							if const.appcfg.debug:
								print(f'Error: {error}')
				else:
					if const.appcfg.debug:
						print(f"No Sync Module for {product}")
			if is_golden:
				model.golden_nets[net_name].networks[
					net_name].lastsync = datetime.utcnow()
				model.golden_nets[net_name].networks[
					net_name].syncruntime = time.perf_counter() - t1
			else:
				
				model.meraki_nets[org_id].networks[
					net_id].lastsync = datetime.utcnow()
				model.meraki_nets[org_id].networks[
					net_id].syncruntime = time.perf_counter() - t1
		if is_golden:
			model.golden_nets[const.appcfg.tag_golden].networks[
				const.appcfg.tag_golden].change_log_ts = await lib.get_network_last_change_ts(
				sdk, org_id, net_id)
		
		print(f'{lib.bc.OKGREEN}Built Config From Network: {net_name} '
		      f'Process took: {net.syncruntime:0.5f} secounds{lib.bc.Default}')
	
	else:
		t = time.perf_counter() - t1
		print(f'Network: {net_name} '
		      f'Skiped Syncing Process took: {t:0.5f} secounds')





async def check_last_change(change, dashboard_change):
	if change != dashboard_change:
		return True
	else:
		return False


async def check_is_golden(tags, net_id):
	if const.appcfg.tag_golden in tags or net_id == 'L_575334852396597314Z':
		return True
	else:
		return False



"""
Module calls the differnt cloning function dynaically based on model values
"""
from meraki.exceptions import AsyncAPIError
from autosync import const, lib, model
from .resync import re_sync


async def clone(org_id: str, net_id: str, sdk: object, product: str):
    """
		Initial Clone Function
	Args:
		org_id: ORG ID
		net_id: Network ID
		sdk: Meraki SDK Object
		product: Current Product for updating

	Returns:

	"""
    golden_tag = const.appcfg.tag_golden
    meraki_function = getattr(sdk, product)
    tasks = (model.golden_nets[golden_tag].networks[golden_tag].
             dashboard[product].settings())
    o_func = model.meraki_nets[org_id].networks[net_id].functions[product]['update']
    net_name = model.meraki_nets[org_id].networks[net_id].name
    for task in tasks:
        if task in o_func:
            try:
                action = model.meraki_nets[org_id].networks[net_id].dashboard[product]
                network_setting = getattr(action, f'Update_{task}')
                await network_setting(sdk,org_id, net_id, net_name)
                # await eval(
                #  f"network.Update_{task}( master,appcfg,sdk,net_id,task,net_name)")
            except AsyncAPIError as error:
                print(f'{lib.bc.FAIL} Network: {net_name} ')
                print(
                    f'\t\t - {lib.bc.WARNING}Override for {product} '
                    f'Task: {task} failed with Error {error}{lib.bc.Default}')
            except Exception as error:
                print(f'{lib.bc.FAIL} Network: {net_name} ')
                print(
                    f'\t\t - {lib.bc.WARNING}Override for {product} '
                    f'Task: {task} failed with Error {error}{lib.bc.Default}')
        else:
            golden_setting = getattr((model.golden_nets[golden_tag].
                                      networks[golden_tag].dashboard[product]),
                                     task)
            network_setting = getattr((
                model.meraki_nets[org_id].networks[net_id].dashboard[product]),
                                      task)
            if not await lib.compare(golden_setting, network_setting):
                print(f'\t {lib.bc.OKGREEN}-Updating {task}...{lib.bc.ENDC}')
                if const.appcfg.write:
                    configure = getattr(meraki_function, f"update{task}")
                    try:
                        await configure(net_id, **golden_setting)
                        await re_sync(sdk, org_id, net_id, task, product)
                        lib.print_update(net_name, task, product)
                    except AsyncAPIError as error:
                        print(
                            f'\t {lib.bc.FAIL} Error Running Setting '
                            f'{task} Error Message: {str(error)}  {lib.bc.Default}'
                        )
                else:
                    print("Write Disabled")
            else:
                lib.print_matched(net_name, task, product)




"""
setup for click
"""
from setuptools import setup,find_packages
req = ['aiohttp','async-timeout','attrs','bcolors',
'certifi','chardet','idna','meraki','multidict','requests',
'typing-extensions','urllib3','yarl','http3','click','pandas',
'tabulate','python-dateutil','automodinit','python-dotenv','wheel',"deepdiff"]
setup(
    name='Meraki_Auto_Sync',
    version='1.182',
    author='Josh Lipton',
    author_email='joliptn@cisco.com',
    description='CLI Tool To Sync Meraki Network Accross Orginizations based on Tags',
	packages=find_packages(),
    packages_data=['autosync'],
    include_package_data=True,
	install_requires=[req],
    entry_points="""
        [console_scripts]
        autosync=autosync.cli.cli:cli
    """,
)



from .bcolors import bcolors as bc
from .cache import clear_cache, load_cache, store_cache
from .find import idFromName, matchGidByName, get_golden_group_policy, get_network_group_policy,find_rf_profile,find_rf_profile_id_by_name
from .helpers import aironetie, rfp_pwr, set_aironet_ie, print_update, print_matched,get_golden, rf_profile_pre_proccess,add_valid_channels,get_change_log_from_org,update_change_log,last_change,get_network_last_change_ts
from .merakiapi import MerakiApi, MerakiAsyncApi
from .netcompare import compare, soft_compare
